package be.jeffcheasey88.peeratcode.parser.state;

import java.util.ArrayList;
import java.util.List;
import java.util.function.BiFunction;
import java.util.function.Function;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.TokenValidator;

public class BuilderStateTree<E, B> extends StateTree<B>{
	
	private BiFunction<E, Bag, B> builder;
	
	private List<StateTree<E>> childs;
	private List<StateTree<E>> multiple;

	public BuilderStateTree(BiFunction<E, Bag, B> builder){
		super();
		this.builder = builder;
		
		this.childs = new ArrayList<>();
		this.multiple = new ArrayList<>();
	}
	
	B build(TokenValidator validator, E element){
		B build = this.builder.apply(element, validator.getBag());
		validator.setBag(new Bag());
		
		BuilderStateTree<B, ?> builded = super.internalSeed(validator, build);
		if(builded != null) builded.build(validator, build);
		
		BuilderStateTree<E, ?> sub_builded;
		
		for(StateTree<E> state : this.multiple){
			TokenValidator branch = validator.branch();
			if(state.checker == null){
				while((sub_builded = state.internalSeed(branch, element)) != null){
					validator.merge(branch);
					sub_builded.build(validator, element);
				}
			}else{
				while(state.checker.apply(branch)){
					sub_builded = state.internalSeed(branch, element);
					if(sub_builded == null) break;
					validator.merge(branch);
					sub_builded.build(validator, element);
				}
			}
		}
		for(StateTree<E> child : this.childs){
			TokenValidator branch = validator.branch();
			if(child.checker == null){
				sub_builded = child.internalSeed(branch, element);
				if(sub_builded == null) continue;
				validator.merge(branch);
				sub_builded.build(validator, element);
				break;
			}
			if(child.checker.apply(branch)){
				sub_builded = child.internalSeed(branch, element);
				if(sub_builded == null) continue;
				validator.merge(branch);
				sub_builded.build(validator, element);
				break;
			}
		}
		
		return build;
	}
	
	public <T extends StateTree<E>> StateTree<E> thenNoChild(StateTree<E> child){
		childs.add(child);
		return child;
	}
	
	public <T extends StateTree<E>> StateTree<E> thenNoChild(Function<TokenValidator, Boolean> checker){
		StateTree<E> state = new StateTree<>();
		state.checker = checker;
		this.childs.add(state);
		return state;
	}
	
	public <T extends StateTree<E>> StateTree<E> multiple(StateTree<E> child){
		this.multiple.add(child);
		return child;
	}
}
