package be.jeffcheasey88.peeratcode.parser.state.generator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import be.jeffcheasey88.peeratcode.parser.Bag;
import be.jeffcheasey88.peeratcode.parser.Parser;
import be.jeffcheasey88.peeratcode.parser.Token;
import be.jeffcheasey88.peeratcode.parser.TokenType;
import be.jeffcheasey88.peeratcode.parser.Tokenizer;
import be.jeffcheasey88.peeratcode.parser.state.RedirectStateTree;
import be.jeffcheasey88.peeratcode.parser.state.StateTree;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.SynthaxeComposant.Composant;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.SynthaxeComposant.Composant.ComposantCall;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.SynthaxeComposant.Composant.LinkedComposant;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.SynthaxeComposant.Composant.MultipleWayComposant;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.SynthaxeComposant.Composant.NamedElement;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.SynthaxeComposant.Composant.Redirection;
import be.jeffcheasey88.peeratcode.parser.state.generator.ParserComposantGenerator.SynthaxeComposant.Composant.Validator;

public class ParserComposantGenerator {
	
	public static void main(String[] args) throws Exception{
		ParserComposantGenerator generator = new ParserComposantGenerator(new BufferedReader(new FileReader(new File("C:\\Users\\jeffc\\eclipse-workspace\\peer-at-code-parseur\\java.pc"))));
		generator.generate();
	}
	
	private BufferedReader reader;
	
	public ParserComposantGenerator(BufferedReader reader){
		this.reader = reader;
	}
	
	public void generate() throws Exception{
		SynthaxeComposant synthax = load();
		Map<String, Set<String>> dependency = synthax.dependency();

		while(!dependency.isEmpty()){
			for(Entry<String, Set<String>> composant : dependency.entrySet()){
				if(composant.getValue().isEmpty()){
					synthax.composants.get(composant.getKey()).build();
					
					dependency.remove(composant.getKey());
					for(Entry<String, Set<String>> entry : dependency.entrySet()) entry.getValue().remove(composant.getKey());
					break;
				}
			}
		}
	}
	
	private SynthaxeComposant load() throws Exception{
		SynthaxeComposant composant = new SynthaxeComposant();
		Parser<Element> parser = new Parser<Element>(){
			{
				StateTree<Element> validator = new StateTree<>();
				StateTree<Element> modifiers = new StateTree<>();
				modifiers.end((composant, bag) -> {
					Element result = new Validator(bag);
					bag.set(result);
					return result;
				});
				StateTree<Element> end = modifiers.then((v) -> v.validate(
						(t) -> t.getValue().equals("="),
						(bag,t) -> bag.set("end", true)));
				StateTree<Element> loop = modifiers.then((v) -> v.validate(
						(t) -> t.getValue().equals("*"),
						(bag,t) -> bag.set("loop", true)));
				loop.then(end);
				loop.end((composant, bag) -> {
					Element result = new Validator(bag);
					bag.set(result);
					return result;
				});
				end.end((composant, bag) -> {
					Element result = new Validator(bag);
					bag.set(result);
					return result;
				});
				
				StateTree<Element> simpleAction = validator
					.then((v) -> v.validate((t) -> t.getValue().equals("[")))
					.then((v) -> v.validate((t) -> !t.getValue().equals("]"), (bag, t) -> bag.set(t))).loop()
					.then((v) -> v.validate((t) -> t.getValue().equals("]"), (bag, t) -> bag.set("type", "simple")));
				StateTree<Element> customAction = validator
						.then((v) -> v.validate((t) -> t.getValue().equals("{")))
						.then((v) -> v.validate((t) -> !t.getValue().equals("}"), (bag, t) -> bag.set(t))).loop()
						.then((v) -> v.validate((t) -> t.getValue().equals("}"), (bag, t) -> bag.set("type", "custom")));
				simpleAction.then(modifiers);
				customAction.then(modifiers);
				
				StateTree<Element> namedValidator = new StateTree<>();
				StateTree<Element> leftSide = namedValidator
					.then((v) -> v.validate((t) -> t.getValue().equals("-")) && v.validate((t) -> t.getValue().equals(">")))
					.then((v) -> v.validate(
							(t) -> t.getType().equals(TokenType.NAME),
							(bag, t) -> bag.set(t)));
				leftSide.end((composant, bag) -> {
					Element result = new NamedElement(bag);
					bag.set(result);
					return result;
				});
				leftSide.then((v) -> 
						v.validate((t) -> t.getValue().equals("(")) &&
						v.validate((t) -> t.getType().equals(TokenType.NAME), (bag, t) -> bag.set("inside", t)) &&
						v.validate((t) -> t.getValue().equals(")"))
						)
						.end((composant, bag) -> {
							Element result = new NamedElement(bag);
							bag.set(result);
							return result;
						});
				
				StateTree<Element> operations = new StateTree<>();
				StateTree<Element> operation = new StateTree<>();
				
				StateTree<Element> sub_operations = new StateTree<>();
				
				StateTree<Element> gotoValidator = new StateTree<>();
				gotoValidator.then((v) -> v.validate(
						(t) -> t.getType().equals(TokenType.NAME),
						(bag, t) -> bag.set(t)))
					.then((v) -> v.validate((t) -> t.getValue().equals(":")))
					.<Element>end((composant, bag) -> {
						Composant parent = (Composant)composant;
						Composant comp = new Composant();
						comp.title = parent.title+"_"+bag.<Token>get().getValue();
						parent.declare(bag.<Token>get().getValue(), comp.current);
						
						parent.currentDeclared = comp;
						return parent;
					}).then(new RedirectStateTree<>(sub_operations, (global, local) -> {
						global.set("operations", local);
					}))
					.end((composant, bag) -> {
						((Composant)composant).currentDeclared = null;
						return composant;
					});
				
				StateTree<Element> redirection = new StateTree<>();
				redirection.then((v) -> v.validate((t) -> 
					 t.getType().equals(TokenType.NAME),
					 (bag, t) -> bag.set(t)
				)).end((composant, bag) -> {
					Element result = new Redirection(bag);
					bag.set(result);
					return result;
				});
				
				StateTree<Element> methodCall = new StateTree<>();
				methodCall.then((v) -> v.validate((t) -> t.getValue().equals(":")))
						.then((v) -> v.validate(
								(t) -> t.getType().equals(TokenType.NAME),
								(bag, t) -> bag.set(t)))
						.end((composant, bag) -> {
							Element result = new ComposantCall(bag);
							bag.set(result);
							return result;
						});
				
				StateTree<Element> op_childs = new StateTree<>();
				op_childs.then((v) -> v.validate((t) -> t.getValue().equals("(")))
					.then(new RedirectStateTree<>(operations, (global, local) -> {
						global.set("operations", local);
					}))
					.then((v) -> v.validate((t) -> t.getValue().equals(")")))
					.end((composant, bag) -> {
						return composant;
					});
				
				operation.then(new RedirectStateTree<>(gotoValidator, (global, local) -> global.set("goto", local))).end((a,b) -> a);
				
				StateTree<Element> op_validator = operation.then(new RedirectStateTree<>(validator, (global, local) -> global.set("validator", local)));
				op_validator.<Element>end((composant, bag) -> {
					Validator result = bag.<Bag>get("validator").<Validator>get();
					bag.set(result);
					Composant comp = new Composant();
					result.composant = comp;
					return comp;
				}).then(op_childs);
				op_validator.then(new RedirectStateTree<>(namedValidator, (global, local) -> global.set("namedValidator", local))).<Element>end((composant, bag) -> {
					NamedElement result = null;
					(result = bag.<Bag>get("namedValidator").<NamedElement>get()).child = bag.<Bag>get("validator").<Validator>get();
					bag.set(result);
					return new Composant();
				}).then(op_childs);
				
				StateTree<Element> op_call = operation.then(new RedirectStateTree<>(methodCall, (global, local) -> global.set("methodCall", local)));
				op_call.<Element>end((composant, bag) -> {
					ComposantCall result = bag.<Bag>get("methodCall").<ComposantCall>get();
					bag.set(result);
					Composant comp = new Composant();
					result.composant = comp;
					return comp;
				}).then(op_childs);
				op_call.then(new RedirectStateTree<>(namedValidator, (global, local) -> global.set("namedValidator", local))).<Element>end((composant, bag) -> {
					NamedElement result = null;
					(result = bag.<Bag>get("namedValidator").<NamedElement>get()).child = bag.<Bag>get("methodCall").<Validator>get();
					bag.set(result);
					return new Composant();
				}).then(op_childs);
				
				StateTree<Element> op_redirect = operation.then(new RedirectStateTree<>(redirection, (global, local) -> global.set("redirection", local)));
				op_redirect.<Element>end((composant, bag) -> {
					Redirection result = bag.<Bag>get("redirection").<Redirection>get();
					bag.set(result);
					Composant comp = new Composant();
					result.composant = comp;
					return comp;
				}).then(op_childs);
				op_redirect.then(new RedirectStateTree<>(namedValidator, (global, local) -> global.set("namedValidator", local))).<Element>end((composant, bag) -> {
					NamedElement result = null;
					(result = bag.<Bag>get("namedValidator").<NamedElement>get()).child = bag.<Bag>get("redirection").<Validator>get();
					bag.set(result);
					return new Composant();
				}).then(op_childs);
				
				StateTree<Element> suit = operations.then(new RedirectStateTree<>(operation, (global, local) ->  {
					Integer count = global.get("count");
					if(count == null) count = 0;
					global.set("count", count+1);
					global.set("operation"+count, local);
				})).loop();
				suit.then((v) -> v.validate((t) -> t.getValue().equals(",")))
					.end((composant, bag) -> {
						Composant current = (Composant)composant;
						if(current.currentDeclared != null) current = current.currentDeclared;
						MultipleWayComposant way = current.current;
						
						int count = bag.get("count");
						LinkedComposant linked = new LinkedComposant();
						for(int i = 0; i < count; i++){
							Bag op = bag.get("operation"+i);
							if(op.get() == null) continue;
							linked.elements.add(op.get());
						}
						way.elements.add(linked);
						return composant;
					}).then(suit);
				suit.then((v) -> v.validate((t) -> t.getValue().equals(";")))
					.end((composant, bag) -> {
						Composant current = (Composant)composant;
						if(current.currentDeclared != null) current = current.currentDeclared;
						MultipleWayComposant way = current.current;
						
						int count = bag.get("count");
						LinkedComposant linked = new LinkedComposant();
						for(int i = 0; i < count; i++){
							Bag op = bag.get("operation"+i);
							if(op.get() == null) continue;
							linked.elements.add(op.get());
						}
						way.elements.add(linked);
						current.include();
						return composant;
					}).then(suit);
				suit.end((composant, bag) -> {
					Composant current = (Composant)composant;
					if(current.currentDeclared != null) current = current.currentDeclared;
					MultipleWayComposant way = current.current;
					
					int count = bag.get("count");
					LinkedComposant linked = new LinkedComposant();
					for(int i = 0; i < count; i++){
						Bag op = bag.get("operation"+i);
						if(op.get() == null) continue;
						linked.elements.add(op.get());
					}
					way.elements.add(linked);
					current.include();
					return composant;
				});
				
				
				StateTree<Element> sub_suit = sub_operations.then(new RedirectStateTree<>(operation, (global, local) ->  {
					Integer count = global.get("count");
					if(count == null) count = 0;
					global.set("count", count+1);
					global.set("operation"+count, local);
				})).loop();
				sub_suit.then((v) -> v.validate((t) -> t.getValue().equals(",")))
					.end((composant, bag) -> {
						Composant current = (Composant)composant;
						if(current.currentDeclared != null) current = current.currentDeclared;
						MultipleWayComposant way = current.current;
						
						int count = bag.get("count");
						LinkedComposant linked = new LinkedComposant();
						for(int i = 0; i < count; i++){
							Bag op = bag.get("operation"+i);
							if(op.get() == null) continue;
							linked.elements.add(op.get());
						}
						way.elements.add(linked);
						return composant;
					}).then(suit);
				sub_suit.then((v) -> v.validate((t) -> t.getValue().equals(";")))
					.end((composant, bag) -> {
						Composant current = (Composant)composant;
						if(current.currentDeclared != null) current = current.currentDeclared;
						MultipleWayComposant way = current.current;
						
						int count = bag.get("count");
						LinkedComposant linked = new LinkedComposant();
						for(int i = 0; i < count; i++){
							Bag op = bag.get("operation"+i);
							if(op.get() == null) continue;
							linked.elements.add(op.get());
						}
						way.elements.add(linked);
						current.include();
						return composant;
					});
				sub_suit.end((composant, bag) -> {
					Composant current = (Composant)composant;
					if(current.currentDeclared != null) current = current.currentDeclared;
					MultipleWayComposant way = current.current;
					
					int count = bag.get("count");
					LinkedComposant linked = new LinkedComposant();
					for(int i = 0; i < count; i++){
						Bag op = bag.get("operation"+i);
						if(op.get() == null) continue;
						linked.elements.add(op.get());
					}
					way.elements.add(linked);
					current.include();
					return composant;
				});
				
				StateTree<Element> main = new StateTree<>();
				
				StateTree<Element> composant = main.then((v) -> v.validate(
						(token) -> token.getType().equals(TokenType.NAME),
						(bag, token) -> bag.set(token)
				));
				composant.<Element>end((element, bag) -> {
					Composant child = new Composant(bag);
					SynthaxeComposant base;
					if(element instanceof SynthaxeComposant){
						base = (SynthaxeComposant)element;
					}else{
						base = ((Composant)element).parent;
					}
					child.parent = base;
					base.register(child.title, child);
					return child;
				}).then(new RedirectStateTree<>(op_childs, (global, local) -> global.set("content", local)))
				.then(composant);
				
				
				setTokenizer(new Tokenizer());
				setStateTree(main);
			}
		};
		
		parser.parse(reader, composant);
//		parser.build(new BufferedWriter(new FileWriter(new File("C:\\Users\\jeffc\\eclipse-workspace\\peer-at-code-parseur\\java.out"))));
		
		return composant;
	}
	
	public static class Element{
		
		public Element(){}
		
		void dependency(Set<String> dependency) {}
	}
	
	public static class SynthaxeComposant extends Element{
		
		private Map<String, Composant> composants;
		
		public SynthaxeComposant(){
			this.composants = new HashMap<>();
		}
		
		void register(String name, Composant composant){
			this.composants.put(name, composant);
		}
		
		private Map<String, Set<String>> dependency(){
			Map<String, Set<String>> map = new HashMap<>();
			
			for(Entry<String, Composant> entry : this.composants.entrySet()){
				Set<String> set = entry.getValue().dependency();
				set.remove(entry.getKey());
				map.put(entry.getKey(), set);
			}
			
			return map;
		}
		
		@Override
		public String toString(){
			String elements = "";
			for(Entry<String, Composant> entry : composants.entrySet()){
				elements+=",\n"+entry.getKey()+":"+entry.getValue();
			}
			if(elements.length() > 1) elements = elements.substring(2);
			return "SynthaxeComposant(\n"+elements+"\n)";
		}
		
		public static class Composant extends Element{
			
			private SynthaxeComposant parent;
			
			private String title;
			
			private MultipleWayComposant path;
			private Map<String, MultipleWayComposant> declared;
			private MultipleWayComposant next;
			
			private MultipleWayComposant current;
			private Composant currentDeclared;
			
			public Composant(){
				this.declared = new HashMap<>();
				
				current = new MultipleWayComposant();
			}
			
			public Composant(Bag bag){
				this();
				this.title = bag.<Token>get().getValue();
			}
			
			public Set<String> dependency(){
				Set<String> result = new HashSet<>();
				if(path != null) path.dependency(result);
				for(MultipleWayComposant way : declared.values()) way.dependency(result);
				if(next != null) next.dependency(result);
				return result;
			}
			
			public void build(){
				System.out.println("build "+title);
				
			}
			
			public void include(){
				if(path == null){
					path = current;
					current = new MultipleWayComposant();
				}else{
					next = current;
					current = new MultipleWayComposant();
				}
			}
			
			public void declare(String value, MultipleWayComposant composant){
				this.declared.put(value, composant);
			}
			
			@Override
			public String toString(){
				return title+"("+path+", "+declared+", "+next+")";
			}
			
			public static class LinkedComposant extends Element{
				
				private List<Element> elements;
				
				public LinkedComposant(){
					this.elements = new ArrayList<>();
				}
				
				@Override
				public void dependency(Set<String> dependency){
					for(Element element : elements) element.dependency(dependency);
				}
				
				
				@Override
				public String toString(){
					return "["+Arrays.toString(elements.toArray())+"]";
				}
			}
			
			public static class MultipleWayComposant extends Element{
				
				private List<Element> elements;
				
				public MultipleWayComposant(){
					this.elements = new ArrayList<>();
				}
				
				@Override
				public void dependency(Set<String> dependency){
					for(Element element : elements) element.dependency(dependency);
				}
				
				@Override
				public String toString(){
					return "<"+Arrays.toString(elements.toArray())+">";
				}
			}
			
				public static class Validator extends Element{
					
					private String action;
					private boolean end;
					private boolean loop;
					
					private Composant composant;
					
					public Validator(Bag bag){
						String type = bag.get("type");
						action = bag.<Token>get().getValue();
						if(type.equals("simple")){
							action = '['+action+']';
						}else{
							action = '{'+action+'}';
						}
						this.end = (bag.get("end") != null);
						this.loop = (bag.get("loop") != null);
					}
					
					@Override
					public void dependency(Set<String> dependency){
						if(composant != null) composant.dependency(dependency);
					}
					
					@Override
					public String toString(){
						return action+((loop ? "*":""))+((end ? "=":""))+((composant == null || composant.path == null) ? "":"("+composant+")");
					}
				}
				
				public static class NamedElement extends Element{
					
					private String logic;
					private String name;
					
					private Element child;
					
					public NamedElement(Bag bag){
						Token inside = bag.get("inside");
						if(inside != null){
							this.logic = bag.<Token>get().getValue();
							this.name = inside.getValue();
						}else{
							this.name = bag.<Token>get().getValue();
						}
					}
					
					@Override
					public void dependency(Set<String> dependency){
						child.dependency(dependency);
					}
					
					@Override
					public String toString(){
						return child+" -> "+((logic == null) ? name : logic+"("+name+")");
					}
					
				}
				
				public static class Redirection extends Element{
					
					private String target;
					
					private Composant composant;
					
					public Redirection(Bag bag){
						this.target = bag.<Token>get().getValue();
					}
					
					@Override
					public void dependency(Set<String> dependency){
						dependency.add(target);
						if(composant != null) composant.dependency(dependency);
					}
					
					@Override
					public String toString(){
						return target+((composant == null || composant.path == null) ? "":"("+composant+")");
					}
				}
				
				public static class ComposantCall extends Element{
					
					private String name;
					
					private Composant composant;
					
					public ComposantCall(Bag bag){
						this.name = bag.<Token>get().getValue();
					}
					
					@Override
					public void dependency(Set<String> dependency){
						if(composant != null) composant.dependency(dependency);
					}
					
					@Override
					public String toString(){
						return ":"+name+((composant == null || composant.path == null) ? "":"("+composant+")");
					}
				}
		}
	}
}
